#!/usr/bin/perl

# Version 1.1
$VERSION=1.1;
# Script to log data Fluke 45 meter
# http://photonicsguy.ca/reference/fluke45

# Serial port to use
# Ex. /dev/ttyUSB0 or /dev/rfcomm0 or /dev/ttyS0
$SERIALPORT='/dev/rfcomm0';

# Record Seperator
$sp=",";
#$sp=":\t";



use Device::SerialPort;
use POSIX qw(strftime);
#use DateTime::Format::Excel;
use Getopt::Std;
use Time::HiRes qw(usleep gettimeofday tv_interval);
use Curses::UI;
use File::Basename;
use IO::Handle;

use constant { TRUE => 1, FALSE => 0 };
#$debug=1;
use warnings;
#use diagnostics;
no warnings 'uninitialized';

binmode STDOUT, ":utf8";
#use Time::HiRes qw(usleep ualarm gettimeofday tv_interval);


$STATUS{'connected'}=FALSE;
$STATUS{'logging'}=FALSE;
$STATUS{'paused'}=FALSE;
$STATUS{'error'}='';

  getopt('s', \%args);

# Prototypes
sub fluke_connect;



#$_=$0;
#m/(.*\/)(.*)/;


my $cui = new Curses::UI(
	-color_support => 1,
	-read_timeout=>0,
	-clear_on_exit=>0,
	-debug => 0,
);
my @menu = (
    { -label => 'File',
        -submenu => [
            { -label => 'Exit      ^Q', -value => \&exit_dialog  },
            { -label => 'Save CSV  ^S', -value => \&save_dialog  },
 #           { -label => 'Close file^ ', -value => \&close_dialog  },
            { -label => 'Pause     ^P', -value => \&pause_dialog  }

        ]},
	{ -label => 'CSV',
		-submenu => [
			{ -label => 'Pause window ^P', -value => \&pause_dialog  },
			{ -label => '          ^', -value => \&pause_dialog  },
			{ -label => '          ^', -value => \&pause_dialog  }

        ]},

	{
	-label => 'Options',
		-submenu => [
			{ -label => 'Led Control', -value => \&light_dialog }
		]
	
    },
);
sub exit_dialog()
{
#    my $return = $cui->dialog(
#        -message   => "Do you really want to quit?",
#        -title     => "Are you sure?",
#        -buttons   => ['yes', 'no'],

#    );
$return=1;
    if($return) {
	print FHCSV "\n";
	close(FHCSV);
	$port->close();
	exit(0);
	}
}
sub save_dialog() {
	if(!$STATUS{'logging'}) {
#		my $mask = [ [ '.',        'All files (*)'       ],[ '\.csv$',   'CSV files (*.csv)'  ] ];    


		if($args{s} ne "") {
			$file=$args{s};
			undef $args{s};
		}else{
			my $fname = strftime("%Y%m%d-%H%M%S", localtime);
#			$fname="charge";
			$file = $cui->savefilebrowser(-file => "$fname.csv",-editfilename => TRUE);# -mask => $mask,-maskselected => 2);
		}

		if(!length $file) { return 0;}


#		if( -e $file ) {			
#		my($fname, $dir, $ext) = fileparse($file);
#		if($file==''){return 0;}
#		if( -d $dir

		if(open(FHCSV, ">$file")) {
			$STATUS{'logging'}=TRUE;
			$STATUS{'paused'}=FALSE;
			$PaneCSV->title("CSV Output ($file)");
			$t0=[gettimeofday];					# Set t0=now()
			CSVline(csvheader());
			FHCSV->autoflush;	# Write data as we go (Useful for using Livegraph)
		} else { $cui->error('Error creating $file'); }
			return 0;
	} else {
		if(!close(FHCSV)) {# || die "Couldn't close file properly";
			$cui->error("Couldn't close file handle");
		} else {
			$PaneCSV->title("CSV Output");
			$STATUS{'logging'}=FALSE;
			my ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,$atime,$mtime,$ctime,$blksize,$blocks) = stat($file);
			$size=sprintf("%.2f",$size/1024);
			push(@csvdata,"\n","Closed $file\n");
			push(@csvdata,"Filesize: $size KBytes\n","\n");
		}
	}
	return 0;
}

sub pause_dialog() {

	if($STATUS{'logging'}) {
		if(!$STATUS{'paused'}) {
			$STATUS{'paused'}=TRUE;
			$PaneCSV->title("CSV Output ($file) PAUSED");
		}else{
			$STATUS{'paused'}=FALSE;
			$PaneCSV->title("CSV Output ($file)");
		}
	}
	return 0;
}

sub light_dialog()
{
    my $return = $cui->dialog(
        -message   => "Toggle Lights",
        -title     => "Yellow",
        -buttons   => ['On', 'Off'],

    );
#	$port->write("$uno,0,".$return."#");
#    exit(0) if $return;
}

my $menu = $cui->add(
    'menu','Menubar',
    -menu => \@menu,
    -fg  => "black",
    -bg => "white"
);
my $screen = $cui->add(
	'screen', 'Window',
	-padtop          => 1, # leave space for the menu
	-border		 => 0,
	-ipad		 => 0,
);


my $PaneLeft = $screen->add(
	'win1', 'Window',
	-border => 1,
	-title=>	"Primary",
	-width=>	$screen->width/2,
	-height=>	10,
	-padtop		 => 0,	
	-showlines	 => 0,
	-sbborder	 => 0,
	-vscrollbar	 => 0,
	-hscrollbar	 => 0,

    -bfg  => 'blue',
);
my $PaneRight = $screen->add(
    'win2', 'Window',
    -border => 1,
	-width=>	$screen->width/2,
	-x=>		$screen->width/2,
	-title=>	"Secondary",
	-padtop		 => 0,	
	-height 	 => 10,
	-showlines	 => 0,
	-sbborder	 => 0,
	-vscrollbar	 => 0,
	-hscrollbar	 => 0,

    -bfg  => 'blue',
);
$PaneCSV = $screen->add(
    'win3', 'Window',
    -border => 1,
	-width=>	$screen->width,
	-x=>		0,
	-y=>		10,
	-title=>	"CSV Output",
	-padtop		 => 0,	
	-height 	 => 10,
	-showlines	 => 0,
	-sbborder	 => 0,
	-vscrollbar	 => 1,
	-hscrollbar	 => 0,

    -bfg  => 'blue',
);
$CSVfile = $PaneCSV->add( 
	'mytextviewer', 'TextViewer',
	-text => ""
    );

$DebugLine = $screen->add(
	'debugline', 'TextEntry',
	-y 	 	 => -2,
	-width		 => -1,
	-reverse 	 => 1,
	-paddingspaces   => 1,
	-readonly =>	1,
	-focusable  => 0,
	-fg              => "blue",
        -bg              => "white",
	-text 	 	 => "---- Line of test text ----",
);

$StatusLine = $screen->add(
	'statusline', 'TextEntry',
	-y 	 	 => -1,
	-width		 => -1,
	-reverse 	 => 1,
	-paddingspaces   => 1,
	-readonly =>	1,
 	-focusable  => 0,
	-fg              => "blue",
        -bg              => "white",
	-text 	 	 => "---- Line of test text ----",
);

$Primary=$PaneLeft->add(
   'llbl'.$i++,'TextEntry',
   -x => ($PaneLeft->width/2)-10,
   -y => 1,
   -width=>$PaneLeft->width/2,
   -readonly => 1,
   -fg => "black",
   -text => '---',
   -focusable  => 0 );

$llbl=$PaneLeft->add(
   'llbl'.$i++,'Label',
   -x => 0,
   -y => $PaneLeft->height-5,
   -width=>6,
   -text => 'Max: ',
);

$PriMax=$PaneLeft->add(
   'llbl'.$i++,'TextEntry',
   -x => 6,
   -y => $PaneLeft->height-5,
   -readonly => 1,
   -fg => "black",
   -text => '---',
   -focusable  => 0 );

$llbl=$PaneLeft->add(
   'llbl'.$i++,'TextEntry',
   -x => 0,
   -y => $PaneLeft->height-4,
   -width=>6,
   -readonly => 1,
   -fg => "black",
   -text => 'Avg: ',
   -focusable  => 0 );

$PriAvg=$PaneLeft->add(
   'llbl'.$i++,'TextEntry',
   -x => 6,
   -y => $PaneLeft->height-4,
   -readonly => 1,
   -fg => "black",
   -text => '---',
   -focusable  => 0 );
$llbl=$PaneLeft->add(
   'llbl'.$i++,'TextEntry',
   -x => 0,
   -y => $PaneLeft->height-3,
   -width=>6,
   -readonly => 1,
   -fg => "black",
   -text => 'Min: ',
   -focusable  => 0 );

$PriMin=$PaneLeft->add(
   'llbl'.$i++,'TextEntry',
   -x => 6,
   -y => $PaneLeft->height-3,
   -readonly => 1,
   -fg => "black",
   -text => '---',
   -focusable  => 0 );

$Secondary=$PaneRight->add(
   'rlbl'.$i,'TextEntry',
   -x => ($PaneRight->width/2)-10,
   -y => 1,
   -width=>$PaneRight->width/2,
   -readonly => 1,
   -fg => "black",
   -text => '---',
   -focusable  => 0 );


$cui->set_binding( sub {$menu->focus()}, "\cX");
$cui->set_binding( \&exit_dialog , "\cQ");
$cui->set_binding( \&save_dialog , "\cS");
$cui->set_binding( sub { $math{'avgcnt'}=0; }, " ");
#$cui->set_binding( \&pause_dialog , "\cP");






#############################################################################
#############################################################################
$i=0;
fluke_connect();
@csvdata=();
$math{'avgcnt'}=0;
$cui->{-read_timeout} = 0.1;
$PaneLeft->focus();

if(defined $args{s}) {
	if($STATUS{'connected'}) {
		save_dialog();
	}else{
		exit 1;
	}
}


$DebugLine->text("");
$cui->add_callback("mainloop", \&mainloop );
$cui->mainloop;

#$cui->do_one_event;
#$cui->draw();

#	 Main LOOP		# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # MAIN
sub mainloop {
#$DebugLine->text("Logging: ".$STATUS{'logging'}."\tPaused:".$STATUS{'paused'}."\t");
	fluke_meas();
	CSVline($csv);
#$cui->draw();
#$PaneCSV->draw();
#$Primary->draw();
#$DebugLine->draw();

$cui->draw();
#$savefilebrowser->draw();
}

#				# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #	

sub fluke_connect() {
	if($port = Device::SerialPort->new($SERIALPORT)) {	# Set up the serial port	9600,8N1 
		#|| die "Can't open $SERIALPORT: $!\n";
		$port->databits(8);
		$port->baudrate(9600);
		$port->parity("none");
		$port->stopbits(1);
		$port->are_match("\n","\r");    # possible end strings<S-Del>
		$port->lookclear;
		$STATUS{'connected'}=TRUE;
		# Attempt to communicate with meter
		($man,$model,$serial,$version)=split(/,/,fluke_command("*idn?"));
		if($man eq "FLUKE") {
			#"Connected.\n";
			fluke_write("FORMAT 2");
		} else {
			$STATUS{'error'}="No response from meter";
			$STATUS{'connected'}=FALSE;
			return 1;
		}
	}else {
		$STATUS{'error'}="Can't open port: $SERIALPORT";
	}
}

sub CSVline {
	if($STATUS{'logging'}) {# and !$STATUS{'paused'}) {
		print FHCSV $_[0]."\n";
		
		foreach(split('\n',$_[0])) { push(@csvdata,$_."\n"); }
	}
	while(@csvdata>$CSVfile->height) { shift(@csvdata) };
	$CSVfile->text(join('',@csvdata));

}

sub fluke_write {
	if(!$STATUS{'connected'}) { return ''; }
	my $cmd=$_[0];
	$port->write($cmd."\r");
	$port->lookclear;
	my $gotit = "";
	while(1) {
		until ("" ne $gotit) {
			$gotit = $port->lookfor;       # poll until data ready
			die "Aborted without match\n" unless (defined $gotit);
		}
		if($gotit eq uc($cmd)) {
			#print "ECHO";
			$gotit="";
		} elsif($gotit eq "=>") {
			#print "PRMPT";
			$gotit="";
			return $gotit;
		}
	}
}


sub fluke_command {
	if(!$STATUS{'connected'}) { return ''; }
	my $cmd=$_[0];
	$port->write($cmd."\r");
	$port->lookclear;
	my $gotit = "";

	my $j=0;
	my $t3=[gettimeofday];

	while(1) {
		until ("" ne $gotit) {
		$j++;
 			$gotit = $port->lookfor;       # poll until data ready
			usleep(50000);
		if($j>300) {	# If this times out, then the meter stopped responding
			$STATUS{'connected'}=FALSE;
			return "+1E9 NOC";

		}
		}
		if($gotit eq uc($cmd)) { $gotit=""; }	# Look for echoed command
		elsif($gotit eq "=>") {
			$gotit="";
my $t4=[gettimeofday];
my $elapsed=tv_interval $t3, $t4;
$DebugLine->text("Serial delay: $elapsed\t\t$j");

			return $value;
		} else {
			$value=$gotit;
			$gotit="";
		}
	}
	return ",,,";
}

sub fluke_meas {
	if(!$STATUS{'connected'}) {
	$StatusLine->text("Not Connected\t".$STATUS{'error'}."\ti:".$i++);	
	usleep(50000);
	$csv='';
	return '';
}
	my $_=fluke_command("meas?");
	my $t1=[gettimeofday];
	$csv='';
	$StatusLine->text($_);

	my ($meas1,$meas2)=split(',',$_);
	$_=humanReadable($meas1);
	$Primary->text(sprintf("%*s",10+length($_)/2,$_));
	if($meas2) {
		$_=humanReadable($meas2);
		$Secondary->text(sprintf("%*s",10+length($_)/2,$_));
	}else { $Secondary->text(""); }
	
	# CSV
	my $elapsed=tv_interval $t0, $t1;
	my ($s, $usec) = values($t1);
	my ($val1,$unit1)=split(' ',$meas1);
	if($val1 eq "+1E+9") { $val1="-";}
	$csv="";
	#$csv="$s.$usec,";
	$csv.=join($sp,$elapsed,$val1,$unit1);

	# Math (Max/Avg/Min)
	if($math{'avgcnt'}==0) {
		$math{'avg'}=0;
		$math{'max'}=$val1-1;
		$math{'min'}=$val1+1;
	}
	# Max
#=comment
	if($math{'max'}<sprintf("%.10f",$val1)) {
		$math{'max'}=$val1;
		$PriMax->text(humanReadable($meas1));
	}
	if($math{'min'}>sprintf("%.10f",$val1)) {
		$math{'min'}=$val1;
		$PriMin->text(humanReadable($meas1));
	}
	$math{'avg'}+=sprintf("%.10f",$val1);
	$math{'avgcnt'}++;
	$avg=$math{'avg'}/$math{'avgcnt'};
	$PriAvg->text(sprintf("%.3f",$avg)." ".$unit1);

#=cut
	if($meas2) {	# Also add second display if enabled
		my ($val2,$unit2)=split(' ',$meas2);
		if($val2 eq "+1E+9") { $val2="-";}
		$csv=join($sp,$csv,$val2,$unit2);
	}

}

sub humanReadable {
	my ($val,$units)=split(' ',$_[0]);
	my $value=scaledunits($val);
	if($units eq "OHMS") {$units="\x{03a9}";}
	elsif($units eq "VDC") {$units="Vdc";}
	elsif($units eq "VAC") {$units="Vac";}
	elsif($units eq "ADC") {$units="Adc";}
	elsif($units eq "AAC") {$units="Aac";}
	elsif($units eq "HZ") {$units="Hz";}
	#elsif($units eq "VDC") {$units="Vdc";}
	#elsif($units eq "VDC") {$units="Vdc";}
	return "$value$units";
}

sub scaledunits { 
	my ($value,$exp)=split('E',$_[0]);
	my (undef,$digits)=split('\.',$value);
	my $precision=length($digits);
	   if($exp==0) { return sprintf("% .*f ",$precision,$value); }
	elsif($exp==-3) { return sprintf("% .*f m",$precision,$value); }
	elsif($exp==3) { return sprintf("% .*f k",$precision,$value); }
	elsif($exp==6) { return sprintf("% .*f M",$precision,$value); }
	elsif($exp==9) { return "OL "; } # Infinite resistance is interpreted as +1E+9
	else { return $_[0]." "; }
}

sub csvheader {
	my ($s,$usec)=values($t0);
	my $date = strftime("%c", localtime($s));
	my $csv="##,##\n#Date: ".$date."\n";
	$csv.="#Port $SERIALPORT\n";
	$csv.="#Model $man$model\n";
	$csv.="#Serial$serial\n";
	$csv.="#Version$version\n";
	$csv.="#Record separator: '$sp'\n";
	$csv.="#T0$sp$s.$usec\n";
	$csv.="#Fluke data logging, Version $VERSION\n#http://photonicsguy.ca/reference/fluke45\n\n";
	$csv.="Time,Primary,P unit,Secondary,S unit";
	return $csv;
}


